/*
 * Code under GPL
 * Author: Patrick "Petschge" Kilian <petschge@web.de>
 * Based on work from: Eric Harlow
 *
 */
#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "bots.h"

// Anzahl der Bots
#ifndef COUNT
#  define COUNT 20
#endif
// Sichtweite^2
#ifndef DISTANCE_S
#  define DISTANCE_S 10000
#endif
// Mindestabstand^2
#ifndef DISTANCE_F
#  define DISTANCE_F 100
#endif
// Breite des Feldes
#ifndef WIDTH
#  define WIDTH 400
#endif
// Hhe des Feldes
#ifndef HEIGHT
#  define HEIGHT 400
#endif
BOT current[COUNT];
BOT next[COUNT];
unsigned int i = 0;
unsigned int j = 0;
signed int dist = 0;
signed int dist_x = 0;
signed int dist_y = 0;
signed int dist_c_x = 0;
signed int dist_c_y = 0;
unsigned long ran = 0;
unsigned int n = 0;

/*
 * init
 *
 * set initial values
 */
void init() {
	srand(time(0));
	for(i=0; i < COUNT; i++){
		current[i].pos_x = rand() % WIDTH;
		current[i].pos_y = rand() % HEIGHT;
		current[i].fly_x = (rand() % 20) - 10;
		current[i].fly_y = (rand() % 20) - 10;
	}
	i = 0;
}

/*
 * draw_bot
 *
 * draw a bot in the given drawing_area, around the given position
 */
void draw_bot (gpointer data, gint x_pos, gint y_pos) {
	GtkWidget*  drawing_area = (GtkWidget *) data;
	GdkDrawable *drawable;

	drawable = drawing_area->window;
	gdk_draw_arc(drawable, drawing_area->style->black_gc, 0, x_pos-2, y_pos-2, 4, 4, 0, 360 * 64);
	gdk_draw_point(drawable, drawing_area->style->black_gc, x_pos, y_pos);
}

/*
 * Repaint
 *
 * data - widget to repaint
 */
gint Repaint (gpointer data) {
	GtkWidget*  drawing_area = (GtkWidget *) data;
	int maxx, maxy;
	GdkDrawable *drawable;

	/* --- Get drawing area window --- */
	drawable = drawing_area->window;

	/* --- Clear rectangle --- */
	gdk_draw_rectangle (drawable, drawing_area->style->white_gc, TRUE, 0, 0,
		drawing_area->allocation.width, drawing_area->allocation.height);

	/* --- Get the size --- */
	maxx = drawing_area->allocation.width;
	maxy = drawing_area->allocation.height;
	
	/* --- Do the drawing --- */
	for(i = 0; i < COUNT; i++) {
		draw_bot ((gpointer) drawing_area, current[i].pos_x, current[i].pos_y);
		// fprintf(stderr, "%d: %d, %d, %d, %d\n", i, current[i].pos_x, current[i].pos_y, current[i].fly_x, current[i].fly_y);
	}

	/* --- Calculate next generation --- */
	for(i = 0; i < COUNT; i++) {
		next[i] = current[i];
		dist_c_x = 0;
		dist_c_y = 0;
		n = 0;
		for(j = 0; j < COUNT; j++) {
			if(! (i == j)) {
				dist_x = (current[j].pos_x - current[i].pos_x);
				dist_y = (current[j].pos_y - current[i].pos_y);
				dist = (dist_x * dist_x) + (dist_y * dist_y);
				if(dist < DISTANCE_S){
					n++;
					/* --- Awoid colisions --- */
					if (dist < DISTANCE_F){
						// fprintf(stderr, "collision: %d <-> %d\n", i, j);
						if( dist_x < 0 ) {
							next[i].pos_x += 3;
							next[i].pos_x += (rand() % 4);
						} else {
							next[i].pos_x -= 3;
							next[i].pos_x -= (rand() % 4);
						}
						if( dist_y < 0 ) {
							next[i].pos_y += 3;
							next[i].pos_x += (rand() % 4);
						} else {
							next[i].pos_y -= 3;
							next[i].pos_x -= (rand() % 4);
						}
					}
					/* --- Align path of flight --- */
						next[i].fly_x += current[j].fly_x;
						next[i].fly_y += current[j].fly_y;
					/* --- Steer towards the folks center --- */
						dist_c_x += dist_x;
						dist_c_y += dist_y;
				}
			}
		}
		if (n) {
			next[i].fly_x /= n+1;
			next[i].fly_y /= n+1;
		}
		next[i].fly_x += dist_c_x / 10;
		next[i].fly_y += dist_c_y / 10;
	}

	for(i = 0; i < COUNT; i++) {
		if (next[i].fly_x > 10) {
			next[i].fly_x = 10;
		}
		if (next[i].fly_x < -10) {
			next[i].fly_x = -10;
		}
		if (next[i].fly_y < -10) {
			next[i].fly_y = -10;
		}
		if (next[i].fly_y > 10) {
			next[i].fly_y = 10;
		}
		current[i].pos_x = next[i].pos_x + next[i].fly_x;
		current[i].pos_y = next[i].pos_y + next[i].fly_y;
		current[i].pos_x %= maxx;
		current[i].pos_y %= maxy;
		current[i].fly_x = next[i].fly_x;
		current[i].fly_y = next[i].fly_y;
	}

	return(TRUE);
}

/*
 * quit
 *
 * exit gtk message loop
 */
void quit () {
	gtk_exit (0);
}


/*
 * main
 *
 * Program begins here.
 */
int main (int argc, char *argv[]) {
	GtkWidget *window;
	GtkWidget *drawing_area;
	GtkWidget *vbox;

	/* --- Initialize GTK --- */
	gtk_init (&argc, &argv);

	/* --- Create a top-level window --- */
	window = gtk_window_new (GTK_WINDOW_TOPLEVEL);

	/* --- Get a packing box --- */
	vbox = gtk_hbox_new (FALSE, 0);

	/* --- Add packing box to window --- */
	gtk_container_add (GTK_CONTAINER (window), vbox);

	/* --- Make packing box visible --- */
	gtk_widget_show (vbox);

	/* --- Listen for signals --- */
	gtk_signal_connect (GTK_OBJECT (window), "destroy", GTK_SIGNAL_FUNC (quit), NULL);

	/* --- Create the drawing area ---*/
	drawing_area = gtk_drawing_area_new ();

	/* --- Set the size --- */
	gtk_drawing_area_size (GTK_DRAWING_AREA (drawing_area), WIDTH, HEIGHT);

	/* --- Add drawing area to packing box --- */
	gtk_box_pack_start (GTK_BOX (vbox), drawing_area, TRUE, TRUE, 0);

	/* --- Make drawing area visible --- */
	gtk_widget_show (drawing_area);

	/* --- Make the window visible --- */
	gtk_widget_show (window);

	/* --- Init the bots --- */
	init();

	/* --- Repaint every 100ms --- */
	gtk_timeout_add (100, Repaint, (gpointer) drawing_area);

	/* --- Start gtk message loop --- */
	gtk_main ();

    return 0;
}
